/*
 *  Driver for HCF softmodem virtual serial ports
 *
 *	Written by Marc Boucher <marc@mbsi.ca>
 */

/*
 * Copyright (c) 2001-2002 Conexant Systems, Inc.
 * 
 * 1.  General Public License. This program is free software, and may
 * be redistributed or modified subject to the terms of the GNU General
 * Public License (version 2) or the GNU Lesser General Public License,
 * or (at your option) any later versions ("Open Source" code). You may
 * obtain a copy of the GNU General Public License at
 * http://www.fsf.org/copyleft/gpl.html and a copy of the GNU Lesser
 * General Public License at http://www.fsf.org/copyleft/less.html,
 * or you may alternatively write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA.
 * 
 * 2.   Disclaimer of Warranties. CONEXANT AND OTHER CONTRIBUTORS MAKE NO
 * REPRESENTATION ABOUT THE SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.
 * IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTIES OF ANY KIND.
 * CONEXANT AND OTHER CONTRIBUTORS DISCLAIMS ALL WARRANTIES WITH REGARD TO
 * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE, GOOD TITLE AND AGAINST INFRINGEMENT.
 * 
 * This software has not been formally tested, and there is no guarantee that
 * it is free of errors including, but not limited to, bugs, defects,
 * interrupted operation, or unexpected results. Any use of this software is
 * at user's own risk.
 * 
 * 3.   No Liability.
 * 
 * (a) Conexant or contributors shall not be responsible for any loss or
 * damage to Company, its customers, or any third parties for any reason
 * whatsoever, and CONEXANT OR CONTRIBUTORS SHALL NOT BE LIABLE FOR ANY
 * ACTUAL, DIRECT, INDIRECT, SPECIAL, PUNITIVE, INCIDENTAL, OR CONSEQUENTIAL
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED, WHETHER IN CONTRACT, STRICT OR OTHER LEGAL THEORY OF
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 * 
 * (b) User agrees to hold Conexant and contributors harmless from any
 * liability, loss, cost, damage or expense, including attorney's fees,
 * as a result of any claims which may be made by any person, including
 * but not limited to User, its agents and employees, its customers, or
 * any third parties that arise out of or result from the manufacture,
 * delivery, actual or alleged ownership, performance, use, operation
 * or possession of the software furnished hereunder, whether such claims
 * are based on negligence, breach of contract, absolute liability or any
 * other legal theory.
 * 
 * 4.   Notices. User hereby agrees not to remove, alter or destroy any
 * copyright, trademark, credits, other proprietary notices or confidential
 * legends placed upon, contained within or associated with the Software,
 * and shall include all such unaltered copyright, trademark, credits,
 * other proprietary notices or confidential legends on or in every copy of
 * the Software.
 * 
 */
#include <linux/version.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/tqueue.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/major.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <asm/serial.h>

#include "oscompat.h"
#include "ostypedefs.h"
#include "comtypes.h"
#include "comctrl_ex.h"
#include "osresour_ex.h"
#include "osservices.h"
#include "oslinux.h"
#ifdef TARGET_HCF_USB_LINUX
#include "usbhalos.h"
#endif

#include "serial_core.h"

#define NR_PORTS		1 // only one HCF port supported for now

#define HCF_ISR_PASS_LIMIT	256

#define HCF_READBUF_SIZE	256

static struct tty_driver hcf_tty_driver_normal;
static struct tty_driver hcf_tty_driver_callout;
static struct tty_struct *hcf_tty_table[NR_PORTS];
static struct termios *hcf_termios[NR_PORTS], *hcf_termios_locked[NR_PORTS];

static PI_COM_CTRL_T hcf_pcomctrl;       // Communication Controller interface
static HANDLE        hcf_hcomctrl;       // Communication Controller instance

static int hcf_rxenabled;
static int hcf_txenabled;

static int hcf_evt_rxchar;
static int hcf_evt_rxbreak;
static int hcf_evt_rxovrn;
static int hcf_evt_txempty;

static unsigned char hcf_readbuf[HCF_READBUF_SIZE];
static int hcf_readcount, hcf_readoffset;

static u_int hcf_mctrl_flags;

static struct tq_struct hcf_intr_tqueue;

static struct uart_info *hcf_uart_info;

#ifdef COMCTRL_MONITOR_POUND_UG_SUPPORTED
static int loglastcallstatus = 0;
#endif

static void
hcf_sched_intr(void)
{
	if(hcf_uart_info) {
		MOD_INC_USE_COUNT;
		if (OsModemThreadSchedule(&hcf_intr_tqueue) == 0) {
			MOD_DEC_USE_COUNT;
		}
	}
}

static void
hcf_stop_tx(struct uart_port *port, u_int from_tty)
{
	//printk(KERN_DEBUG "%s\n", __FUNCTION__);
	hcf_txenabled = 0;
}

static void
hcf_start_tx(struct uart_port *port, u_int nonempty, u_int from_tty)
{
	//printk(KERN_DEBUG "%s\n", __FUNCTION__);
	hcf_txenabled = 1;
	if (nonempty) {
		hcf_sched_intr();
	}
}

static void
hcf_stop_rx(struct uart_port *port)
{
	//printk(KERN_DEBUG "%s\n", __FUNCTION__);
	hcf_rxenabled = 0;
}

static inline int
hcf_rx_ready(void)
{
    int r;

	if(!hcf_rxenabled)
		return FALSE;

    if(hcf_readoffset == hcf_readcount) {
		r = hcf_pcomctrl->Read(hcf_hcomctrl, hcf_readbuf, sizeof(hcf_readbuf));
        if(r < 0) {
			printk(KERN_ERR"%s: ComCtrlRead returned %d\n", __FUNCTION__, r);
		} else {
			hcf_readcount = r;
			hcf_readoffset = 0;
		}
	}

    return (hcf_readcount > hcf_readoffset) || hcf_evt_rxbreak || hcf_evt_rxovrn;
}

static inline void
hcf_rx_chars(struct uart_info *info)
{
	struct tty_struct *tty = info->tty;
	struct uart_port *port = info->port;
	int max_count = sizeof(hcf_readbuf);

	hcf_evt_rxchar = 0;

	while(max_count-- > 0 && hcf_rx_ready()) {
		if(tty->flip.count >= TTY_FLIPBUF_SIZE) {
			tty->flip.tqueue.routine((void *) tty);
			if(tty->flip.count >= TTY_FLIPBUF_SIZE) {
				return; // if TTY_DONT_FLIP is set
			}
		}

		if (hcf_evt_rxovrn) {
	    	hcf_evt_rxovrn = 0;
	    	port->icount.overrun++;

			*tty->flip.flag_buf_ptr++ = TTY_OVERRUN;
			*tty->flip.char_buf_ptr++ = 0;
			tty->flip.count++;
			continue;
		}

		port->icount.rx++;

		if (hcf_evt_rxbreak) {
	    	hcf_evt_rxbreak = 0;
	    	port->icount.brk++;

			*tty->flip.flag_buf_ptr++ = TTY_BREAK;
		} else
			*tty->flip.flag_buf_ptr++ = TTY_NORMAL;

		*tty->flip.char_buf_ptr++ = hcf_readbuf[hcf_readoffset++];
		tty->flip.count++;
	}

	tty_flip_buffer_push(tty);
	return;
}

static inline int
hcf_tx_free(void)
{
	int r;
	UINT32 val = 0;

    if ((r=hcf_pcomctrl->Monitor(hcf_hcomctrl, COMCTRL_MONITOR_TXFREE, &val)) != COM_STATUS_SUCCESS) {
		printk(KERN_ERR "%s: ComCtrlMonitor COMCTRL_MONITOR_TXFREE failed, status=%d\n", __FUNCTION__, r);
		return 0;
	}

	//printk(KERN_DEBUG "%s: val=%lu\n", __FUNCTION__, val);
	return val;
}

static inline void
hcf_put_char(struct uart_port *port, unsigned char ch)
{
    int r;

	//printk(KERN_DEBUG "%s: ch=%x\n", __FUNCTION__, (int)ch);
	hcf_evt_txempty = 0;
    r = hcf_pcomctrl->Write(hcf_hcomctrl, &ch, 1);
    if(r != 1) {
		printk(KERN_ERR "%s: ComCtrlWrite returned %d\n", __FUNCTION__, r);
    }
}

static inline void
hcf_tx_chars(struct uart_info *info)
{
	struct uart_port *port = info->port;

	if (port->x_char) {
		hcf_put_char(port, port->x_char);
		port->icount.tx++;
		port->x_char = 0;
		return;
	}
	if (info->xmit.head == info->xmit.tail
	    || info->tty->stopped
	    || info->tty->hw_stopped) {
		hcf_stop_tx(info->port, 0);
		return;
	}

	while (hcf_tx_free() > 0) {
		hcf_put_char(port, info->xmit.buf[info->xmit.tail]);
		info->xmit.tail = (info->xmit.tail + 1) & (UART_XMIT_SIZE - 1);
		port->icount.tx++;
		if (info->xmit.head == info->xmit.tail)
			break;
	}

	if (CIRC_CNT(info->xmit.head, info->xmit.tail, UART_XMIT_SIZE) <
			WAKEUP_CHARS)
		uart_event(info, EVT_WRITE_WAKEUP);

	if (info->xmit.head == info->xmit.tail)
		hcf_stop_tx(info->port, 0);
}

static inline int
hcf_tx_ready(void) {
	return (hcf_txenabled /*&& hcf_evt_txempty*/);
}

static void
hcf_intr(void *dev_id)
{
	unsigned int pass_counter = 0;

	while (hcf_uart_info && (hcf_rx_ready() || hcf_tx_ready())) {
		hcf_rx_chars(hcf_uart_info);

		if(hcf_tx_ready())
			hcf_tx_chars(hcf_uart_info);

		if (pass_counter++ > HCF_ISR_PASS_LIMIT)
			break;
	}

	MOD_DEC_USE_COUNT;
}

/*
 * Return TIOCSER_TEMT when transmitter is not busy.
 */
static u_int
hcf_tx_empty(struct uart_port *port)
{
	return hcf_evt_txempty ? TIOCSER_TEMT : 0;
}

static u_int
hcf_get_mctrl(struct uart_port *port)
{
	return hcf_mctrl_flags;
}

#ifdef COMCTRL_MONITOR_POUND_UG_SUPPORTED
static COM_STATUS
hcf_monitor(COMCTRL_MONITOR_CODE eCode, PVOID pMonitor)
{
	int r;

	if(!hcf_pcomctrl || !hcf_hcomctrl)
		return COM_STATUS_INST_NOT_INIT;

   	r = hcf_pcomctrl->Monitor(hcf_hcomctrl, eCode, pMonitor);

    if (r != COM_STATUS_SUCCESS) {
		printk(KERN_ERR "%s: ComCtrlMonitor %d failed, status=%d\n", __FUNCTION__, eCode, r);
	}

	return r;
}
#endif

static COM_STATUS
hcf_control(COMCTRL_CONTROL_CODE eCode, PVOID pControl)
{
	int r;

	if(!hcf_pcomctrl || !hcf_hcomctrl)
		return COM_STATUS_INST_NOT_INIT;

   	r = hcf_pcomctrl->Control(hcf_hcomctrl, eCode, pControl);

    if (r != COM_STATUS_SUCCESS) {
		printk(KERN_ERR "%s: ComCtrlControl %d failed, status=%d\n", __FUNCTION__, eCode, r);
	}

	return r;
}

static void
hcf_set_mctrl(struct uart_port *port, u_int mctrl)
{
	//printk(KERN_DEBUG "%s: mctrl=%x\n", __FUNCTION__, mctrl);

	if ((mctrl & TIOCM_RTS) && !(hcf_mctrl_flags & TIOCM_RTS)) {
		hcf_mctrl_flags |= TIOCM_RTS;
		hcf_control(COMCTRL_CONTROL_SETRTS, 0);
	}
	if (!(mctrl & TIOCM_RTS) && (hcf_mctrl_flags & TIOCM_RTS)) {
		hcf_mctrl_flags &= ~TIOCM_RTS;
		hcf_control(COMCTRL_CONTROL_CLRRTS, 0);
	}

	if ((mctrl & TIOCM_DTR) && !(hcf_mctrl_flags & TIOCM_DTR)) {
		hcf_mctrl_flags |= TIOCM_DTR;
		hcf_control(COMCTRL_CONTROL_SETDTR, 0);
	}
	if (!(mctrl & TIOCM_DTR) && (hcf_mctrl_flags & TIOCM_DTR)) {
		hcf_mctrl_flags &= ~TIOCM_DTR;
		hcf_control(COMCTRL_CONTROL_CLRDTR, 0);
	}
}

static void
hcf_break_ctl(struct uart_port *port, int break_state)
{
	//printk(KERN_DEBUG "%s: break_state=%d\n", __FUNCTION__, break_state);
#if 0
	if(break_state)
		printk(KERN_WARNING "%s: modem does not presently support sending/receiving BREAK\n", __FUNCTION__);
#endif
}

static void
hcf_event_handler(struct uart_port *port, UINT32 dwEvtMask)
{
	u_int mctrl_flags, orig_mctrl_flags;
	int sched_intr=0;

	orig_mctrl_flags = mctrl_flags = hcf_mctrl_flags;
   	//printk(KERN_DEBUG "%s: dwEvtMask=0x%04lx\n", __FUNCTION__, dwEvtMask);

	if((dwEvtMask & COMCTRL_EVT_RXCHAR)) {
	    hcf_evt_rxchar = 1;
	    sched_intr = 1;
	}

	if(dwEvtMask & COMCTRL_EVT_BREAK) {
	    hcf_evt_rxbreak = 1;
	    sched_intr = 1;
	}

	if((dwEvtMask & COMCTRL_EVT_RXOVRN)) {
	    hcf_evt_rxovrn = 1;
	    sched_intr = 1;
	}

#if 0
	if((dwEvtMask & COMCTRL_EVT_TXCHAR)) {
	}
#endif

	if((dwEvtMask & COMCTRL_EVT_TXEMPTY)) {
	    hcf_evt_txempty = 1;
	    sched_intr = 1;
	}

	if((dwEvtMask & COMCTRL_EVT_CTS)) {
		if(dwEvtMask & COMCTRL_EVT_CTSS) {
		    mctrl_flags |= TIOCM_CTS;
		} else {
		    mctrl_flags &= ~TIOCM_CTS;
		}
	}

	if(dwEvtMask & COMCTRL_EVT_DSR) {
		if(dwEvtMask & COMCTRL_EVT_DSRS) {
		    mctrl_flags |= TIOCM_DSR;
		} else {
		    mctrl_flags &= ~TIOCM_DSR;
		}
		port->icount.dsr++;
	}

	if(dwEvtMask & COMCTRL_EVT_RLSD) {
		if(dwEvtMask & COMCTRL_EVT_RLSDS) {
		    mctrl_flags |= TIOCM_CAR;
		} else {
		    mctrl_flags &= ~TIOCM_CAR;
		}
	}

	if(dwEvtMask & COMCTRL_EVT_RING) {
		if(dwEvtMask & COMCTRL_EVT_RINGS) {
		   	mctrl_flags |= TIOCM_RNG;
			port->icount.rng++;
		} else {
			mctrl_flags &= ~TIOCM_RNG;
	    }
	}

	if(hcf_mctrl_flags != mctrl_flags) {
		hcf_mctrl_flags = mctrl_flags;
#if 0
		printk(KERN_DEBUG "%cCTS %cDSR %cDCD %cRI\n",
				hcf_mctrl_flags&TIOCM_CTS?'+':'-',
				hcf_mctrl_flags&TIOCM_DSR?'+':'-',
				hcf_mctrl_flags&TIOCM_CAR?'+':'-',
				hcf_mctrl_flags&TIOCM_RNG?'+':'-');
#endif

		if(hcf_uart_info) {
			if((mctrl_flags & TIOCM_CAR) != (orig_mctrl_flags & TIOCM_CAR)) {
				uart_handle_dcd_change(hcf_uart_info, mctrl_flags & TIOCM_CAR);
#ifdef COMCTRL_MONITOR_POUND_UG_SUPPORTED
				if(loglastcallstatus && !(mctrl_flags & TIOCM_CAR)) {
					PORT_MONITOR_DATA monitorData;
					static char largebuf[PAGE_SIZE];
					char *p, *nl;

					monitorData.dwSize = sizeof(largebuf);
					monitorData.pBuf = largebuf;

					hcf_monitor(COMCTRL_MONITOR_POUND_UG, &monitorData);
					p = largebuf;
					while((nl = strchr(p, '\n'))) {
						printk(KERN_INFO "HCF: %.*s", (nl - p) + 1, p);
						p = nl + 1;
					}
					if(*p)
						printk(KERN_INFO "HCF: %s\n", p);

				}
#endif
			}
			if((mctrl_flags & TIOCM_CTS) != (orig_mctrl_flags & TIOCM_CTS))
				uart_handle_cts_change(hcf_uart_info, mctrl_flags & TIOCM_CTS);

			wake_up_interruptible(&hcf_uart_info->delta_msr_wait);
		}
	}

	if(sched_intr) {
		hcf_sched_intr();
	}

	return;
}

static int
hcf_comctrl_init(void)
{
	int r;

	//printk(KERN_DEBUG "%s\n", __FUNCTION__);

	hcf_pcomctrl = ComCtrlGetInterface();
	if(!hcf_pcomctrl) {
 		printk(KERN_ERR "%s: ComCtrlGetInterface error\n", __FUNCTION__);
		return -EIO;
	}

	if((r=hcf_pcomctrl->GetInterfaceVersion()) != COMCTRL_VERSION) {
   		printk(KERN_DEBUG "%s: ComCtrlGetInterfaceVersion returned 0x%d (expecting %d)\n", __FUNCTION__, r, COMCTRL_VERSION);
		hcf_pcomctrl = NULL;
		return -EIO;
	}

	hcf_hcomctrl = hcf_pcomctrl->Create();
	if(!hcf_hcomctrl) {
 		printk(KERN_DEBUG "%s: ComCtrlCreate failed!\n", __FUNCTION__);
		hcf_hcomctrl = NULL;
		hcf_pcomctrl = NULL;
		return -EIO;
	}
	return 0;
}

static void
hcf_comctrl_exit(void)
{
	if(hcf_pcomctrl && hcf_hcomctrl) {
		hcf_pcomctrl->Destroy(hcf_hcomctrl);

		hcf_pcomctrl = NULL;
		hcf_hcomctrl = NULL;
	}
}

static int
hcf_startup(struct uart_port *port, struct uart_info *info)
{
	//printk(KERN_DEBUG "%s\n", __FUNCTION__);
	if(!hcf_pcomctrl || !hcf_hcomctrl)
		return -ENODEV;

	if(hcf_uart_info)
		return -EBUSY;

	/* flush any characters or events received while we were shutdown */
	while(hcf_pcomctrl->Read(hcf_hcomctrl, hcf_readbuf, sizeof(hcf_readbuf)) > 0);
	hcf_readcount = hcf_readoffset = 0;
	hcf_evt_rxchar = 0;
	hcf_evt_rxbreak = 0;
	hcf_evt_rxovrn = 0;

	hcf_uart_info = info;

	hcf_rxenabled = 1;
	hcf_txenabled = 1;

	return 0;
}

static void
hcf_shutdown(struct uart_port *port, struct uart_info *info)
{
	//printk(KERN_DEBUG "%s\n", __FUNCTION__);
	hcf_rxenabled = 0;
	hcf_txenabled = 0;

	hcf_uart_info = NULL;
}

static void
hcf_change_speed(struct uart_port *port, u_int cflag, u_int iflag, u_int quot)
{
	PORT_CONFIG port_config;

	//printk(KERN_DEBUG "%s\n", __FUNCTION__);

	memset(&port_config, 0, sizeof(port_config));

	if(quot) {
		port_config.dwDteSpeed = port->uartclk / (16 * quot);
		port_config.dwValidFileds |= PC_DTE_SPEED;
	}

	if(cflag & PARENB) {
		if(cflag & PARODD)
			port_config.eParity = PC_PARITY_ODD;
		else
			port_config.eParity = PC_PARITY_EVEN;
	} else
			port_config.eParity = PC_PARITY_NONE;
	port_config.dwValidFileds |= PC_PARITY;

	if((cflag & CSIZE) == CS7) {
		port_config.eDataBits = PC_DATABITS_7;
	} else {
		port_config.eDataBits = PC_DATABITS_8;
	}
	port_config.dwValidFileds |= PC_DATA_BITS;

	if (cflag & CRTSCTS) {
		port_config.fCTS = TRUE;
		port_config.fRTS = TRUE;
	}
	port_config.dwValidFileds |= PC_CTS | PC_RTS;

	hcf_control(COMCTRL_CONTROL_PORTCONFIG, &port_config);
}

static void
hcf_enable_ms(struct uart_port *port)
{
}

static void
hcf_release_port(struct uart_port *port)
{
	//printk(KERN_DEBUG "%s\n", __FUNCTION__);
	hcf_mctrl_flags &= ~TIOCM_DSR; // until comctrl generates events for this

	if(hcf_pcomctrl && hcf_hcomctrl) {
		hcf_pcomctrl->Close(hcf_hcomctrl);
	}
}

static void *hcf_devnode;

static int
hcf_request_port(struct uart_port *port)
{
	PORT_EVENT_HANDLER EvtHandler;
	int r;

	//printk(KERN_DEBUG "%s\n", __FUNCTION__);
	hcf_mctrl_flags = 0;

	hcf_evt_rxchar = 0;
	hcf_evt_rxbreak = 0;
	hcf_evt_rxovrn = 0;
	hcf_evt_txempty = 1;

	hcf_rxenabled = 0;
	hcf_txenabled = 0;

	hcf_readcount = hcf_readoffset = 0;

	INIT_TQUEUE(&hcf_intr_tqueue, hcf_intr, port);

	if ((r=hcf_pcomctrl->Configure(hcf_hcomctrl, COMCTRL_CONFIG_DEVICE_ID, hcf_devnode))) {
 		printk(KERN_DEBUG "%s: ComCtrlConfigure DEVICE_ID failed, r=%d!\n", __FUNCTION__, r);
		return -EIO;
	}

	EvtHandler.pfnCallback = (void (*) (PVOID pRef, UINT32 dwEventMask))
	   				hcf_event_handler;
	EvtHandler.pRef = port;

	if ((r=hcf_pcomctrl->Configure(hcf_hcomctrl, COMCTRL_CONFIG_EVENT_HANDLER, &EvtHandler))) {
 		printk(KERN_DEBUG "%s: ComCtrlConfigure EVENT_HANDLER failed, r=%d!\n", __FUNCTION__, r);
		return -EIO;
	}

	if((r=hcf_pcomctrl->Open(hcf_hcomctrl))) {
 		printk(KERN_ERR "%s: ComCtrlOpen failed, r=%d!\n", __FUNCTION__, r);
		return -EIO;
	}

	hcf_mctrl_flags |= TIOCM_DSR; // until comctrl generates events for this

    return 0;
}

static void
hcf_config_port(struct uart_port *port, int flags)
{
	//printk(KERN_DEBUG "%s\n", __FUNCTION__);
	if (flags & UART_CONFIG_TYPE && hcf_request_port(port) == 0) {
		port->type = PORT_VIRTUAL;
	}
}

/*
 * Verify the new serial_struct (for TIOCSSERIAL).
 * The only change we allow are to the flags
 */
static int
hcf_verify_port(struct uart_port *port, struct serial_struct *ser)
{
	int ret = 0;
	if (ser->type != PORT_VIRTUAL)
		ret = -EINVAL;
	if (port->irq != ser->irq)
		ret = -EINVAL;
	/*if (ser->io_type != SERIAL_IO_MEM)
		ret = -EINVAL;*/
	if (port->uartclk / 16 != ser->baud_base)
		ret = -EINVAL;
	/*if ((void *)port->mapbase != ser->iomem_base)
		ret = -EINVAL;*/
	if (port->iobase != ser->port)
		ret = -EINVAL;
	if (ser->hub6 != 0)
		ret = -EINVAL;
	return ret;
}

static struct uart_ops hcf_pops = {
	tx_empty:		hcf_tx_empty,
	set_mctrl:		hcf_set_mctrl,
	get_mctrl:		hcf_get_mctrl,
	stop_tx:		hcf_stop_tx,
	start_tx:		hcf_start_tx,
	stop_rx:		hcf_stop_rx,
	enable_ms:		hcf_enable_ms,
	break_ctl:		hcf_break_ctl,
	startup:		hcf_startup,
	shutdown:		hcf_shutdown,
	change_speed:	hcf_change_speed,
	release_port:	hcf_release_port,
	request_port:	hcf_request_port,
	config_port:	hcf_config_port,
	verify_port:	hcf_verify_port,
};

static struct uart_port hcf_ports[NR_PORTS] = {
	{
		iobase:		COMCTRL_VERSION,
		uartclk:	BASE_BAUD * 16,
		fifosize:	16,
		ops:		&hcf_pops,
		flags:      ASYNC_BOOT_AUTOCONF,
	},
};

static int serialmajor = HCFSERIALMAJOR;
MODULE_PARM(serialmajor, "i");
MODULE_PARM_DESC(serialmajor, "Major device number for serial device");
#ifdef HCFCALOUTMAJOR
static int calloutmajor = HCFCALOUTMAJOR;
MODULE_PARM(calloutmajor, "i");
MODULE_PARM_DESC(calloutmajor, "Major device number for callout device");
#endif

#ifdef TARGET_HCF_PCI_LINUX
static int pcivendorid = 0;
MODULE_PARM(pcivendorid, "i");
MODULE_PARM_DESC(pcivendorid, "PCI Vendor ID of HCF modem");
static int pcideviceid = 0;
MODULE_PARM(pcideviceid, "i");
MODULE_PARM_DESC(pcideviceid, "PCI Device ID of HCF modem");
#endif

#ifdef COMCTRL_MONITOR_POUND_UG_SUPPORTED
MODULE_PARM(loglastcallstatus, "i");
MODULE_PARM_DESC(loglastcallstatus, "Log AT#UG command output after each connection");
#endif

static struct uart_driver hcf_reg = {
	owner:				THIS_MODULE,
#ifdef CONFIG_DEVFS_FS
	normal_name:		"ttySHCF%d",
	callout_name:		"cuaHCF%d",
#else
	normal_name:		"ttySHCF",
	callout_name:		"cuaHCF",
#endif
	normal_driver:		&hcf_tty_driver_normal,
	callout_driver:		&hcf_tty_driver_callout,
	table:				hcf_tty_table,
	termios:			hcf_termios,
	termios_locked:		hcf_termios_locked,
	minor:				HCFSERIALMINOR,
	nr:					NR_PORTS,
	port:				hcf_ports,
};

#ifndef CNXTSERIAL_INCLUDE_CORE
#define uart_init() 0
#define uart_exit() {}
#endif

#if defined(COMCTRL_MONITOR_POUND_UG_SUPPORTED) && defined(CONFIG_PROC_FS) && (LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0))
static struct proc_dir_entry *hcf_proc_dir;
static struct proc_dir_entry *hcf_proc_unit_dir;
static struct proc_dir_entry *hcf_proc_lastcallstatus;

static int hcf_get_lastcallstatus(char *buf, char **start, off_t offset, int length)
{
	PORT_MONITOR_DATA monitorData;

	monitorData.dwSize = PAGE_SIZE;
	monitorData.pBuf = buf;

	if (hcf_monitor(COMCTRL_MONITOR_POUND_UG, &monitorData) == COM_STATUS_SUCCESS)
		return strlen(buf);
	else
		return -EIO;
}
#endif

static int __init
hcf_serial_init(void)
{
	int ret;

	hcf_reg.normal_major = serialmajor;
	hcf_reg.callout_major = calloutmajor;

	if(serialmajor == 0) {
		printk(KERN_ERR "%s: serialmajor parameter must be non-null\n", __FUNCTION__);
		return -EINVAL;
	}

	if(calloutmajor == 0) {
		printk(KERN_ERR "%s: calloutmajor parameter must be non-null\n", __FUNCTION__);
		return -EINVAL;
	}

	if(serialmajor == calloutmajor) {
		printk(KERN_ERR "%s: serialmajor and calloutmajor parameter values must differ\n", __FUNCTION__);
		return -EINVAL;
	}

#ifdef TARGET_HCF_PCI_LINUX
	if((pcivendorid && !pcideviceid) || (pcideviceid && !pcivendorid)) {
		printk(KERN_ERR "%s: both pcideviceid and pcivendorid parameters must be specified\n", __FUNCTION__);
		return -EINVAL;
	}

	if(pcivendorid && pcideviceid) {
		hcf_devnode = (void *)OSFindPCIDevice(pcivendorid, pcideviceid);
		if(!hcf_devnode) {
			printk(KERN_ERR "%s: PCI device %04x:%04x not found\n", __FUNCTION__, pcivendorid, pcideviceid);
			return -EINVAL;
		}
	} else
		hcf_devnode = (void *)-1; // let devicemanager find modem
#elif defined(TARGET_HCF_USB_LINUX)
	{
		int i;

		for (i=0; i<5; i++) {
			if ((hcf_devnode = OsUsbGetInstance(0)) != NULL)
				break;
			OsSleep(1000);
		}
		if(hcf_devnode == NULL)
			return -ENODEV;
	}
#else
#error Unsupported target!
#endif

	if ((ret = hcf_comctrl_init()) != 0)
		return ret;

	if ((ret = uart_init()) != 0) {
		hcf_comctrl_exit();
		return ret;
	}

	ret = uart_register_driver(&hcf_reg);
	if(ret) {
		uart_exit();
		hcf_comctrl_exit();
		return ret;
	}

#if defined(COMCTRL_MONITOR_POUND_UG_SUPPORTED) && defined(CONFIG_PROC_FS) && (LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0))
	hcf_proc_dir = proc_mkdir("hcf", proc_root_driver);
	if(hcf_proc_dir) {
		hcf_proc_unit_dir = proc_mkdir("0", hcf_proc_dir);
		if(hcf_proc_unit_dir) {
			hcf_proc_lastcallstatus = create_proc_info_entry("lastcallstatus", 0, hcf_proc_unit_dir, hcf_get_lastcallstatus);
		}
	}
#endif

	return ret;
}

static void __exit
hcf_serial_exit(void)
{
#if defined(COMCTRL_MONITOR_POUND_UG_SUPPORTED) && defined(CONFIG_PROC_FS) && (LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0))
	if(hcf_proc_lastcallstatus)
		remove_proc_entry("lastcallstatus", hcf_proc_unit_dir);
	if(hcf_proc_unit_dir)
		remove_proc_entry("0", hcf_proc_dir);
	if(hcf_proc_dir)
		remove_proc_entry("hcf", proc_root_driver);
#endif
	uart_unregister_driver(&hcf_reg);
	uart_exit();
	hcf_comctrl_exit();
}

module_init(hcf_serial_init);
module_exit(hcf_serial_exit);

EXPORT_NO_SYMBOLS;

MODULE_AUTHOR("Marc Boucher <marc@mbsi.ca> for Conexant Systems Inc.");
MODULE_DESCRIPTION("HCF softmodem serial port driver");
MODULE_LICENSE("GPL");

#ifdef CNXTSERIAL_INCLUDE_CORE

#undef EXPORT_SYMBOL
#define EXPORT_SYMBOL(x)

#undef MODULE_DESCRIPTION
#define MODULE_DESCRIPTION(x)

#undef MODULE_LICENSE
#define MODULE_LICENSE(x)

#undef module_init
#define module_init(x)

#undef module_exit
#define module_exit(x)

#include "serial_core.c"
#endif

